# standard imports
from typing import Type
from io import BytesIO
import os
from urllib.parse import urlparse

# package imports
from multimethod import multidispatch

# local imports
from ..settings import (
    SUPPORTS_CSV,
    SUPPORTS_JSON,
    SUPPORTS_PICKLE,
    SUPPORTS_PDF,
)

# Conditional imports dependent on supported file types
if SUPPORTS_PDF:
    from PyPDF2 import PdfReader, PdfWriter
if SUPPORTS_JSON:
    import json
if SUPPORTS_PICKLE:
    import dill


def get_file_path_extension(file_path: str) -> str:
    """
    Get the extension of the file path.
    """
    return file_path.split(".")[-1]


def get_url_extension(url: str) -> str:
    """
    Get the extension of the URL.
    """
    # Parse the URL to get the path component
    parsed_url = urlparse(url)
    # Extract the file path
    file_path = parsed_url.path
    # Get just the filename from the path
    filename = os.path.basename(file_path)
    # Get the extension without the dot (e.g., "txt")
    _, extension = os.path.splitext(filename)
    return extension.replace(".", "")


def get_type(key: str) -> Type:
    """
    Get the type of the file based on the file extension.
    """
    if key.startswith("http"):
        file_type = get_url_extension(key)
    else:
        file_type = get_file_path_extension(key)
    match file_type:
        case "txt":
            return str
        case "csv":
            return str
        case "json":
            return dict
        case "pkl":
            return object
        case "pdf":
            return PdfReader


@multidispatch
def convert(data, return_type):
    raise NotImplementedError(
        f"Automatic conversion from {data.__class__} to {return_type} is not yet supported."
    )


if SUPPORTS_CSV:

    # Conversion to bytes
    @convert.register
    def _(data: str, return_type: Type[bytes]):
        # Convert CSV (string) to bytes
        return data.encode("utf-8")

    # Conversion from bytes
    @convert.register
    def _(data: bytes, return_type: Type[str]):
        # Convert bytes to CSV (string)
        return data.decode("utf-8")


if SUPPORTS_JSON:

    @convert.register
    def _(data: dict, return_type: Type[bytes]):
        # Convert JSON (dict) to bytes
        return json.dumps(data).encode("utf-8")

    @convert.register
    def _(data: bytes, return_type: Type[dict]):
        # Convert bytes to JSON (dict)
        return json.loads(data.decode("utf-8"))


if SUPPORTS_PICKLE:

    @convert.register
    def _(data: bytes, return_type: Type[object]):
        # Pickle is already in bytes, return as is
        return data

    @convert.register
    def _(data: object, return_type: Type[bytes]):
        # Convert object to bytes using pickle
        buffer = BytesIO()
        dill.dump(data, buffer)
        buffer.seek(0)
        return buffer.getvalue()


if SUPPORTS_PDF:

    @convert.register
    def _(data, return_type: Type[bytes]):
        # Convert PdfReader to bytes
        output_buffer = BytesIO()
        pdf_writer = PdfWriter()

        # Copy all pages from the reader to the writer
        for page in data.pages:
            pdf_writer.add_page(page)

        # Copy metadata
        if data.metadata:
            pdf_writer.add_metadata(data.metadata)

        # Write to buffer
        pdf_writer.write(output_buffer)
        output_buffer.seek(0)
        return output_buffer.getvalue()

    @convert.register
    def _(data: bytes, return_type: Type[PdfReader]):
        # Convert bytes to PdfReader
        buffer = BytesIO(data)
        return PdfReader(buffer)



from abc import ABC, abstractmethod
from typing import Any, Optional, Tuple

from pydantic import BaseModel


class StorageConfig(BaseModel):
    pass


class StorageBase(ABC):

    @abstractmethod
    def ping() -> dict:
        pass

    def create_file(
        self, key: str, data: Any, meta_data: Optional[dict] = None
    ) -> None:
        """
        Save a file.
        The data must be of a supported type.
        The meta data, if provided, must be a dictionary.
        If the key is `example/data.csv`, the meta data will be saved to `example/meta.json`.

        Args:
            key (str): The path to the file.
            data (Any): The data to save.
            meta_data (Optional[dict], optional): The meta data to save. Defaults

        Returns:
            None
        """
        pass

    def read_file(self, key: str) -> Tuple[Any, Optional[dict]]:
        """
        Read a file.
        The data will be converted to the appropriate type.

        Args:
            key (str): The path to the file.

        Returns:
            Tuple[Any, Optional[dict]]: The data and, if available, the metadata.
        """
        pass

    def update_file(
        self, key: str, data: Any, meta_data: Optional[dict] = None
    ) -> None:
        """
        Update a file and its metadata.

        Args:
            key (str): The path to the file.
            data (Any): The data to save.

        Returns:
            None

        """
        pass

    def delete_file(self, key: str) -> None:
        """
        Delete a file.

        Args:
            key (str): The path to the file.

        Returns:
            None
        """

    @abstractmethod
    def list_files_in_directory(key: str) -> list:
        pass

    @abstractmethod
    def list_subdirectories_in_directory(key: str) -> list:
        pass



from typing import Optional, Tuple, Any
import requests

import boto3
from botocore.exceptions import ClientError
from mypy_boto3_s3.client import S3Client

from ..base import StorageBase, StorageConfig
from ..convert import convert, get_type
from .. import utils


class S3TmpStorageConfig(StorageConfig):
    bucket: str


class S3TmpStorage(StorageBase):
    """
    A CRUD interface for S3Tmp.
    """

    def __init__(self, config: S3TmpStorageConfig) -> None:
        self.client: S3Client = boto3.client("s3")
        self.bucket = config.bucket

    def ping(self) -> dict:
        try:
            self.client.head_bucket(Bucket=self.bucket)
            return {"response": "pong"}
        except ClientError as e:
            raise e

    def create_file(
        self, key: str, data: Any, meta_data: Optional[dict] = None
    ) -> Tuple[str, Optional[str]]:
        """
        Save a file to S3Tmp.
        The data must be of a supported type.
        The meta data, if provided, must be a dictionary.
        If the key is `example/data.csv`, the meta data will be saved to `example/meta.json`.

        Args:
            key (str): The path to the file.
            data (Any): The data to save.
            meta_data (Optional[dict], optional): The meta data to save. Defaults

        Returns:
            None
        """

        # Save the data
        try:
            data = convert(data, bytes)
            self.client.put_object(Body=data, Bucket=self.bucket, Key=key)
        except Exception as e:
            raise e

        # Save the metadata
        if meta_data is not None:
            try:
                meta_data_key = utils.get_meta_data_key(key)
                meta_data = convert(meta_data, bytes)
                self.client.put_object(
                    Body=meta_data, Bucket=self.bucket, Key=meta_data_key
                )
                meta_data_url = self.client.generate_presigned_url(
                    "get_object", Params={"Bucket": self.bucket, "Key": meta_data_key}
                )
            except Exception as e:
                raise e
        else:
            meta_data_url = None

        # Generate and return a presigned URL
        data_url = self.client.generate_presigned_url(
            "get_object", Params={"Bucket": self.bucket, "Key": key}
        )
        return data_url, meta_data_url

    def read_file(
        self, key: str, meta_data_key: Optional[str] = None
    ) -> Tuple[Any, Optional[dict]]:
        """
        Read a file from S3Tmp.
        The data will be converted to the appropriate type.
        The meta data, if provided, will be converted to a dictionary.

        Args:
            key (str): The path to the file.

        Returns:
            Tuple[Any, Optional[dict]]: The data and, if available, the meta data.
        """

        # Get the data
        try:
            response = requests.get(key)
            data = response.content
            data = convert(data, get_type(key))
        except Exception as e:
            raise e

        # Get the metadata
        if meta_data_key is not None:
            try:
                response = requests.get(meta_data_key)
                meta_data = response.content
                meta_data = convert(meta_data, dict)
            except Exception as e:
                raise e
        else:
            meta_data = None

        return data, meta_data

    def update_file(self, key: str, data: Any, meta_data: Optional[dict]) -> None:
        pass

    def delete_file(self, key: str) -> None:
        pass

    def list_files_in_directory(self, key: str) -> list:
        pass

    def list_subdirectories_in_directory(self, key) -> list:
        pass



import os
import io
from typing import Optional, Tuple, Any
import tempfile

from dotenv import load_dotenv
from google.oauth2.credentials import Credentials
from googleapiclient.discovery import build
from googleapiclient.http import MediaFileUpload, MediaIoBaseDownload

from ..base import StorageBase, StorageConfig
from ..convert import convert, get_type
from .. import utils


# Load environment variables
load_dotenv()
CLIENT_ID = os.getenv("CLIENT_ID")
CLIENT_SECRET = os.getenv("CLIENT_SECRET")
TOKEN_URI = "https://oauth2.googleapis.com/token.json"


class GoogleDriveStorageConfig(StorageConfig):
    # TODO: Should all of these values be here?
    client_id: str = CLIENT_ID
    client_secret: str = CLIENT_SECRET
    token: str
    token_uri: str = TOKEN_URI
    root_folder_id: Optional[str] = None


class GoogleDriveStorage(StorageBase):
    def __init__(self, config: GoogleDriveStorageConfig) -> None:
        """
        Initialize Google Drive API client using OAuth credentials.

        Args:
            storage_config (dict): Configuration dictionary containing:
                {
                    "storage_type": "google_drive",
                    "storage_config": {
                        "token": "...",
                        "token_uri": "...",
                        "root_folder_id": "..." # Optional
                    }
                }
        """

        # Initialize root_folder_id
        self.root_folder_id = config.root_folder_id

        # Create credentials from the token
        self.creds = Credentials(
            client_id=config.client_id,
            client_secret=config.client_secret,
            token=config.token,
            token_uri=config.token_uri,
            scopes=["https://www.googleapis.com/auth/drive.file"],
        )

        self.service = build("drive", "v3", credentials=self.creds)

    def ping(self) -> dict:
        return {"response": "pong"}

    def create_file(
        self, key: str, data: Any, meta_data: Optional[dict] = None
    ) -> None:
        """
        Save a file to Google Drive.
        The data must be of a supported type.
        The meta data, if provided, must be a dictionary.
        If the key is `example/data.csv`, the meta data will be saved to `example/meta.json`.

        Args:
            key (str): The path to the file.
            data (Any): The data to save.
            meta_data (Optional[dict], optional): The meta data to save.

        Returns:
            None
        """
        # Create a temporary file to store the data
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            # Convert data to bytes and write to temp file
            byte_data = convert(data, bytes)
            temp_file.write(byte_data)
            temp_key = temp_file.name

        try:
            # Prepare file metadata
            file_metadata = {"name": os.path.basename(key)}
            if self.root_folder_id:
                file_metadata["parents"] = [self.root_folder_id]

            # Upload the main file
            media = MediaFileUpload(temp_key, resumable=True)
            self.service.files().create(
                body=file_metadata, media_body=media, fields="id"
            ).execute()

            # Handle metadata if provided
            if meta_data is not None:
                meta_data_key = utils.get_meta_data_key(key)
                meta_file_metadata = {"name": os.path.basename(meta_data_key)}
                if self.root_folder_id:
                    meta_file_metadata["parents"] = [self.root_folder_id]

                # Create temporary file for metadata
                with tempfile.NamedTemporaryFile(delete=False) as meta_temp_file:
                    meta_bytes = convert(meta_data, bytes)
                    meta_temp_file.write(meta_bytes)
                    meta_temp_path = meta_temp_file.name

                try:
                    meta_media = MediaFileUpload(meta_temp_path, resumable=True)
                    self.service.files().create(
                        body=meta_file_metadata, media_body=meta_media, fields="id"
                    ).execute()
                finally:
                    os.unlink(meta_temp_path)

        finally:
            os.unlink(temp_key)

    def read_file(self, key: str) -> Tuple[Any, Optional[dict]]:
        """
        Read a file from Google Drive.
        The data will be converted to the appropriate type.
        The meta data, if provided, will be converted to a dictionary.

        Args:
            key (str): The path to the file.

        Returns:
            Tuple[Any, Optional[dict]]: The data and, if available, the meta data.
        """
        # Find the file by name
        file_name = os.path.basename(key)
        response = (
            self.service.files()
            .list(q=f"name = '{file_name}'", spaces="drive", fields="files(id, name)")
            .execute()
        )

        if not response["files"]:
            raise FileNotFoundError(f"File {key} not found in Google Drive")

        file_id = response["files"][0]["id"]

        # Read the main file
        request = self.service.files().get_media(fileId=file_id)
        fh = io.BytesIO()
        downloader = MediaIoBaseDownload(fh, request)

        done = False
        while done is False:
            _, done = downloader.next_chunk()

        # Convert the data to the appropriate type
        data = convert(fh.getvalue(), get_type(key))

        # Try to get metadata
        meta_data = None
        try:
            meta_data_key = utils.get_meta_data_key(key)
            meta_file_name = os.path.basename(meta_data_key)

            response = (
                self.service.files()
                .list(
                    q=f"name = '{meta_file_name}'",
                    spaces="drive",
                    fields="files(id, name)",
                )
                .execute()
            )

            if response["files"]:
                meta_file_id = response["files"][0]["id"]
                request = self.service.files().get_media(fileId=meta_file_id)
                meta_fh = io.BytesIO()
                downloader = MediaIoBaseDownload(meta_fh, request)

                done = False
                while done is False:
                    _, done = downloader.next_chunk()

                meta_data = convert(meta_fh.getvalue(), dict)

        except Exception as e:
            print(f"No metadata found: {str(e)}")
            meta_data = None

        return data, meta_data

    def delete_file(self, key: str) -> None:
        """
        Delete a file and its metadata from Google Drive.
        If the key is `example/data.csv`, both `example/data.csv` and `example/meta.json` will be deleted if they exist.

        Args:
            key (str): The path to the file to delete.

        Returns:
            None
        """
        # Find and delete the main file
        file_name = os.path.basename(key)
        response = (
            self.service.files()
            .list(q=f"name = '{file_name}'", spaces="drive", fields="files(id, name)")
            .execute()
        )

        if not response["files"]:
            raise FileNotFoundError(f"File {key} not found in Google Drive")

        file_id = response["files"][0]["id"]
        self.service.files().delete(fileId=file_id).execute()

        # Try to find and delete the metadata file if it exists
        try:
            meta_data_key = utils.get_meta_data_key(key)
            meta_file_name = os.path.basename(meta_data_key)

            response = (
                self.service.files()
                .list(
                    q=f"name = '{meta_file_name}'",
                    spaces="drive",
                    fields="files(id, name)",
                )
                .execute()
            )

            if response["files"]:
                meta_file_id = response["files"][0]["id"]
                self.service.files().delete(fileId=meta_file_id).execute()

        except Exception as e:
            print(f"No metadata file found to delete: {str(e)}")

    def list_files_in_directory(self, key: str) -> list:
        pass

    def list_subdirectories_in_directory(self, key) -> list:
        pass



from importlib import import_module
from stringcase import pascalcase, snakecase
from typing import Any, Dict

from acrud.storage.base import StorageBase, StorageConfig


def get_config_from_str(storage_type: str, config: Dict[str, Any]) -> StorageConfig:
    package = "acrud.storage"
    try:
        module = import_module(package + "." + storage_type, package)
        storage_config_class = getattr(
            module, f"{pascalcase(storage_type)}StorageConfig"
        )
        return storage_config_class(**config)
    except (ImportError, AttributeError) as e:
        raise ValueError(f"Unsupported storage type: {storage_type}") from e


def create_storage(config: StorageConfig) -> StorageBase:
    package = "acrud.storage"
    storage_type = snakecase(config.__class__.__name__.replace("StorageConfig", ""))
    # Dynamically import the appropriate storage module
    try:
        module = import_module(package + "." + snakecase(storage_type), package)
        storage_class = getattr(module, f"{pascalcase(storage_type)}Storage")
        return storage_class(config)
    except (ImportError, AttributeError) as e:
        raise ValueError(f"Unsupported storage type: {storage_type}") from e



from .storage.base import StorageBase


def lookup_handler(storage: StorageBase, file_path: str):

    # Split the file path by '/'
    split_file_path = file_path.split("/")

    for index, path in enumerate(split_file_path):

        if path == "":
            pass

        short_file_path = "/".join(split_file_path[: index + 1])

        try:
            folders = storage.list_subdirectories_in_directory(short_file_path)
        except FileNotFoundError:
            raise LookupError(f"File not found. No such directory as {short_file_path}")

    raise LookupError(f"File not found.")



import os
from typing import Optional, Tuple, Any

from ..base import StorageBase, StorageConfig
from ..convert import convert, get_type
from ...exception import lookup_handler
from .. import utils


class LocalStorageConfig(StorageConfig):
    root: str


class LocalStorage(StorageBase):
    """
    A CRUD interface for local storage.
    """

    def __init__(self, config: LocalStorageConfig) -> None:
        self.root_dir = config.root

    def ping(self) -> dict:
        return {"response": "pong"}

    def create_file(
        self, key: str, data: Any, meta_data: Optional[dict] = None
    ) -> None:
        """
        Save a file to local storage.
        The data must be of a supported type.
        The meta data, if provided, must be a dictionary.
        If the key is `example/data.csv`, the meta data will be saved to `example/meta.json`.

        Args:
            key (str): The path to the file.
            data (Any): The data to save.
            meta_data (Optional[dict], optional): The meta data to save. Defaults

        Returns:
            None
        """

        key = os.path.join(self.root_dir, key)

        folder = "/".join(key.split("/")[:-1])

        if not os.path.exists(folder):
            os.makedirs(folder)

        # Save the data
        data = convert(data, bytes)
        with open(key, "wb") as f:
            f.write(data)

        # Save the metadata
        if meta_data is not None:
            meta_data_key = utils.get_meta_data_key(key)
            meta_data = convert(meta_data, bytes)
            with open(meta_data_key, "wb") as f:
                f.write(meta_data)

    def read_file(self, key: str) -> Tuple[Any, Optional[dict]]:
        """
        Read a file from local storage.
        The data will be converted to the appropriate type.

        Args:
            key (str): The path to the file.

        Returns:
            Tuple[Any, Optional[dict]]: The data and, if available, the metadata.
        """

        key = os.path.join(self.root_dir, key)

        try:
            with open(key, "rb") as f:
                obj = f.read()
        except FileNotFoundError:
            lookup_handler(self, key)

        data = convert(obj, get_type(key))  # Converts file data

        # If a metadata file exists, read it
        meta_data_key = utils.get_meta_data_key(key)
        if os.path.exists(meta_data_key):
            with open(meta_data_key, "rb") as f:
                obj = f.read()
            meta_data = convert(obj, dict)
        else:
            meta_data = None

        return data, meta_data

    def update_file(
        self, key: str, data: Any, meta_data: Optional[dict] = None
    ) -> None:
        """
        TODO: Implement this method.
        Replace the data in a file in S3.

        Args:
            key (str): The path to the file.
            data (Any): The data to save.

        Returns:
            None

        """

        # In Local we simply overwrite the file
        self.create_file(key, data, meta_data)

    def delete_file(self, key: str) -> None:

        full_key = os.path.join(self.root_dir, key)

        # Delete the data
        try:
            os.remove(full_key)
        except LookupError:
            lookup_handler(self, key)

        # Delete the metadata
        meta_data_key = utils.get_meta_data_key(full_key)
        if os.path.exists(meta_data_key):
            os.remove(meta_data_key)

    def list_files_in_directory(self, key: str) -> list:
        full_path = os.path.join(self.root_dir, key)
        if not os.path.exists(full_path):
            return []

        files = os.listdir(full_path)
        # Remove file extensions and metadata files
        files = [
            os.path.splitext(f)[0]
            for f in files
            if not f.startswith(".") and not f.endswith(".meta.json")
        ]
        return list(set(files))

    def list_subdirectories_in_directory(self, key) -> list:

        path = os.path.join(self.root_dir, key)
        files = os.listdir(path)

        # Filter out entries that start with a dot
        files = [file for file in files if not file.startswith(".")]

        return files



import os
from typing import Optional, Tuple, Any

import boto3
from botocore.exceptions import ClientError
from pydantic import BaseModel

from ..base import StorageBase, StorageConfig
from ..convert import convert, get_type
from .. import utils


class S3StorageConfig(StorageConfig):
    bucket: str


class S3Storage(StorageBase):
    """
    A CRUD interface for S3.
    """

    def __init__(self, config: S3StorageConfig) -> None:
        self.client = boto3.client("s3")
        self.bucket = config.bucket

    def ping(self) -> dict:
        try:
            self.client.head_bucket(Bucket=self.bucket)
            return {"response": "pong"}
        except ClientError as e:
            raise e

    def create_file(
        self, key: str, data: Any, meta_data: Optional[dict] = None
    ) -> None:
        """
        Save a file to S3.
        The data must be of a supported type.
        The meta data, if provided, must be a dictionary.
        If the key is `example/data.csv`, the meta data will be saved to `example/meta.json`.

        Args:
            key (str): The path to the file.
            data (Any): The data to save.
            meta_data (Optional[dict], optional): The meta data to save. Defaults

        Returns:
            None
        """

        # Save the data
        try:
            data = convert(data, bytes)
            self.client.put_object(Body=data, Bucket=self.bucket, Key=key)
        except Exception as e:
            raise e

        # Save the metadata
        if meta_data is not None:
            try:
                meta_data_key = utils.get_meta_data_key(key)
                meta_data = convert(meta_data, bytes)
                self.client.put_object(
                    Body=meta_data, Bucket=self.bucket, Key=meta_data_key
                )
            except Exception as e:
                raise e

    def read_file(self, key: str) -> Tuple[Any, Optional[dict]]:
        """
        Read a file from S3.
        The data will be converted to the appropriate type.
        The meta data, if provided, will be converted to a dictionary.

        Args:
            key (str): The path to the file.

        Returns:
            Tuple[Any, Optional[dict]]: The data and, if available, the meta data.
        """

        # Get the data
        obj = self.client.get_object(Bucket=self.bucket, Key=key)
        obj = obj["Body"].read()
        data = convert(obj, get_type(key))  # Converts file data

        # Get the metadata
        try:
            meta_data_key = utils.get_meta_data_key(key)
            obj = self.client.get_object(Bucket=self.bucket, Key=meta_data_key)
            meta_data = obj["Body"]

            if meta_data is not None:
                meta_data = convert(meta_data.read(), dict)

        except ClientError:
            print("No metadata found.")
            meta_data = None

        # Create the file object
        return data, meta_data

    def update_file(self, key: str, data: Any, meta_data: Optional[dict]) -> None:
        """
        TODO: Implement this method.
        Replace the data in a file in S3.

        Args:
            key (str): The path to the file.
            data (Any): The data to save.

        Returns:
            None

        """

        # In s3 we simply overwrite the file
        self.create_file(key, data, meta_data)

    def delete_file(self, key: str) -> None:

        # Delete the data
        self.client.delete_object(Bucket=self.bucket, Key=key)

        # Delete the metadata
        meta_data_key = utils.get_meta_data_key(key)
        self.client.delete_object(Bucket=self.bucket, Key=meta_data_key)

    def list_files_in_directory(self, key: str) -> list:
        files = self.client.list_objects_v2(Bucket=self.bucket, Prefix=key)
        return files

    def list_subdirectories_in_directory(self, key) -> list:
        def _extract_next_directory(key: str, path: str) -> str:
            directories, _ = os.path.split(key)
            directory_list = directories.split("/")
            directory_index = directory_list.index(path)
            # Check if the directory is the last one in the list
            if directory_index == len(directory_list) - 1:
                return None
            else:
                next_directory = directory_list[directory_index + 1]
                return next_directory

        response = self.client.list_objects_v2(Bucket=self.bucket, Prefix=key)

        last = key.split("/")[-1]
        dirs = []

        files = response.get("Contents")
        if files is not None:
            for file in files:
                next_dir = _extract_next_directory(file["Key"], last)
                if next_dir is not None:
                    dirs.append(next_dir)
            dirs = list(set(dirs))  # Remove duplicates

        return dirs
